import pyglet
import pyglet.window.key as KEY
import cocos
from cocos import path
from cocos.actions import *
from cocos.layer.base_layers import *
from gamelib.ccreader import *
from cocos.scenes import *
import os
import random
import configparser

screen_center = (480, 320)
GAME_WAITING = 0
GAME_START = 1
GAME_OVER = 2
GAME_PAUSED = 3

class Hero:
    def __init__(self, scene):
        attack_sprite_names = ["punch1", "punch2", "kick1", "kick2", "punch3", "punch4"]
        fall_sprite_names = ["fall"]
        stand_sprite_names = ["stand"]
        self.scene = scene
        self.attack_frames = []
        self.attack_index = -1
        for sprite_name in attack_sprite_names:
            node = scene.node_dict[sprite_name][0]
            node.visible = False
            self.attack_frames.append(node)

        self.fall_frames = []
        node = scene.node_dict["fall"][0]
        node.visible = False
        self.fall_frames.append(node)

        self.stand_frames = []
        node = scene.node_dict["stand"][0]
        self.stand_frames.append(node)

        self.hero_node = scene.node_dict["hero"][0]

    def attack(self, scale = 1):
        if self.stand_frames[0].visible:
            self.stand_frames[0].visible = False
            self.attack_index += 1            
        else:
            self.attack_frames[self.attack_index].visible = False
            self.attack_index = (self.attack_index + 1) % len(self.attack_frames)
        frame = self.attack_frames[self.attack_index]
        frame.visible = True
        frame.scale_x = scale

    def fall(self, scale = 1):
        self.attack_frames[self.attack_index].visible = False
        frame = self.fall_frames[0]
        frame.scale_x = scale
        frame.visible = True

        #HACK: to change the z-order of the object. 
        self.scene.remove(self.hero_node)
        self.scene.add(self.hero_node, 10)

        if scale > 0:
            fall_anim = frame.anim_dict["fall"]
        else:
            fall_anim = frame.anim_dict["fall_left"]
        frame.do(fall_anim)
        


class GamePlayScreen(Layer):
    is_event_handler = True
    def __init__(self, scene):
        super(GamePlayScreen, self).__init__()
        self.scene_done = False
        scene.add(self)
        
        #load the sound effects for the game. 
        attack_sound_files = ["punch.wav", "punch1.wav", "punch2.wav"]
        self.attack_sounds = []
        for sf in attack_sound_files:
            self.attack_sounds.append(pyglet.media.load(os.path.join("Resources", "creator", "Audio", sf), streaming=False))
        self.hurt_sound = pyglet.media.load(os.path.join("Resources", "creator", "Audio","attack.wav"), streaming=False)

        images = ["walk1.png", "walk2.png"]
        self.skeleton_walk_imgs = list(map(lambda x: pyglet.resource.image(os.path.join("Texture", "skeleton", x)), images))

        images = ["attack1.png", "attack2.png"]
        self.skeleton_attk_imgs = list(map(lambda x: pyglet.resource.image(os.path.join("Texture", "skeleton", x)), images))

        self.skeleton_fall_img = pyglet.resource.image(os.path.join("Texture", "skeleton", "fall.png"))

        #load combat animation images
        self.label = scene.node_dict['label'][0]
        self.label.element.text = "0"
        self.progress_bar = scene.node_dict['pb'][0]

        #HACK to set the foreground objects at the top
        fg = scene.node_dict['foreground'][0]
        scene.remove(fg)
        self.add(fg, 12)

        #HACK figure out how to solve these issues :()
        self.progress_bar.position = (screen_center[0]/2+140, 600-8)
        self.progress_bar.anchor_x= 0
        self.progress_bar.anchor_y = 0
        self.progress_bar.scale_x = 0.5
        
        self.game_score = 0
        self.game_state = GAME_WAITING
        #rate of decrement == 0.1 then 0.05 per punch is essentially 2 punches per second
        self.decrease_rate = 0.05
        self.increment_rate = 0.05
        
        #variables to calculate the new high punches per second. 
        self.time_accum = 0
        self.punches_accum = 0
        self.punches_ps = 0

        #set up the scene. 
        self.hero = Hero(scene)
        self.hero_sprite = scene.node_dict["hero"][0]
        
        #HACK: multiply to reduce the space between enemies and hero.
        GamePlayScreen.spr_width = self.hero.stand_frames[0].width * 0.50

        self.enemies_left = []
        self.enemies_right = []      
        
        for i in range(6):
            self.add_enemy()
        self.print_enemy_lists()

        self.schedule(self.update)

    def handle_success(self):
        self.game_score += 1
        self.label.element.text = str(self.game_score)

        self.punches_accum += 1
        pb_scale = self.progress_bar.scale_x + self.increment_rate
        if pb_scale > 1:
            pb_scale = 1
        self.progress_bar.scale_x = pb_scale

    def handle_game_over(self):
        e = self.enemies_left[0]
        scale = 1
        if not(e):
            e = self.enemies_right[0]
            scale = -1
        e.image = self.skeleton_attk_imgs[1]
        #write the current score into the config
        config = configparser.ConfigParser()
        try: 
            config.read('score.ini')
            if config.has_option("best", "score"):
                val = int(config.get("best","score"))
                if val < self.game_score:
                    config["best"]["score"] = self.game_score
            else:
                config["best"]["score"] = self.game_score
        except:
            #config["best"]["score"] = self.game_score
            if not(config.has_section("best")):                
                config.add_section("best")
            config.set("best","score", str(self.game_score))
        config.write(open("score.ini", "w"))

        self.hurt_sound.play()
        self.hero.fall(scale)
        self.do(sequence(Delay(2), CallFunc(self.switch_2_main)))

    def switch_2_main(self):
        import main
        scn = main.CreateScene()
        cocos.director.director.replace(FadeTransition(scn, duration=0.5))
        print("Not Implemented yet")

    def update(self, dt):
        if self.game_state != GAME_START:
            return

        pb_scale = self.progress_bar.scale_x - self.decrease_rate * dt
        if pb_scale < 0:
            pb_scale = 0
            self.game_state = GAME_OVER        
        self.progress_bar.scale_x = pb_scale

        if self.game_state == GAME_OVER:
            self.handle_game_over()

        #calculate the punches per second using accumulators.
        #TODO: create an animation which highlights the punches per second when it reaches new heights. 
        self.time_accum += dt
        if self.time_accum > 1.0:
            self.time_accum = 0
            ps = self.punches_accum
            self.punches_accum = 0
            if (ps == 2 or self.game_score >= 10) and self.decrease_rate < 0.1:  #TODO: add score criterion
                self.decrease_rate = 0.1
            elif self.game_score >= 25 and self.decrease_rate < 0.1125: #TODO: use this as a final stage beyond 50/25 score.
                self.decrease_rate = 0.1125

    def add_enemy(self):
        is_right = random.choice([0, 1])
        spr = cocos.sprite.Sprite(os.path.join("Texture", "skeleton", "walk1.png"))
        spr.anchor_y = 0
        # spr.color = random.choice([(255, 0, 0), (0, 255, 0), (0, 0, 255)])
        self.add(spr)

        if is_right:
            self.enemies_left.append(None)
            self.enemies_right.append(spr)
            spr.position = (screen_center[0] + self.spr_width * (len(self.enemies_right)), self.hero_sprite.position[1])
            spr.scale_x = -1
        else:
            self.enemies_left.append(spr)
            self.enemies_right.append(None)
            spr.position = (screen_center[0] - self.spr_width * (len(self.enemies_left)), self.hero_sprite.position[1])

    def move_enemies(self):
        move_duration = 0.15
        move_delta = self.spr_width
        index = 1
        for (el, er) in zip(self.enemies_left, self.enemies_right):
            if el:                
                pos = (screen_center[0] - self.spr_width * index, self.hero_sprite.position[1])
                el.do(MoveTo(pos, move_duration))
                el.image = self.skeleton_walk_imgs[index%len(self.skeleton_walk_imgs)]
            elif er:
                pos = (screen_center[0] + self.spr_width * index, self.hero_sprite.position[1])
                er.do(MoveTo(pos, move_duration))
                er.image = self.skeleton_walk_imgs[index%len(self.skeleton_walk_imgs)]
            index += 1

        #pick the closest enemy and set the attack anim. 
        en = self.enemies_left[0]
        if not(en):
            en = self.enemies_right[0]
        en.image = self.skeleton_attk_imgs[0]

    def enemy_fall(self, enemy, is_right=False):
        #HACK to get the z-order to change for the falling enemy.
        self.remove(enemy)
        self.add(enemy, 10)
        enemy.image = self.skeleton_fall_img
        random.choice(self.attack_sounds).play()

        #path.Bezier(a, b, ac, bc)
        if is_right:
            h = random.choice([50, 100, 150, 200])
            bp = path.Bezier((0, 0), (400, 0), (100, h), (400, h))
            enemy.do(spawn(FadeOut(0.5), sequence(Bezier(bp, 0.5), CallFunc(self.remove, enemy))))
        else:
            h = random.choice([50, 100, 150, 200])
            bp = path.Bezier((0, 0), (-400, 0), (-100, h), (-400, h))
            enemy.do(spawn(FadeOut(0.5), sequence(Bezier(bp, 0.5), CallFunc(self.remove, enemy))))

    def print_enemy_lists(self):
        return
        print( list(map(lambda x: x != None, self.enemies_left)) )
        print( list(map(lambda x: x != None, self.enemies_right)) )

    def on_key_press(self, key, modifiers):
        if self.game_state == GAME_OVER:
            return

        if self.game_state == GAME_WAITING and key == KEY.LEFT or key == KEY.RIGHT:
            self.game_state = GAME_START

        if KEY.LEFT == key:
            self.hero.attack()
            # self.label.element.text = "Left"
            if self.enemies_left[0] == None:
                print("Wrong Move")
                #self.move_enemies()
                self.game_state = GAME_OVER
                self.handle_game_over()
            else:
                #pop the first element and add new elements to the back. 
                node = self.enemies_left.pop(0)
                self.enemies_right.pop(0)
                self.enemy_fall(node)
                self.move_enemies()
                self.add_enemy()
                self.handle_success()
        elif KEY.RIGHT == key:
            self.hero.attack(-1)
            # self.label.element.text = "Right"
            if self.enemies_right[0] == None:
                print("Wrong Move")
                self.game_state = GAME_OVER
                self.handle_game_over()                
            else:
                #pop the first element and add new elements to the back. 
                self.enemies_left.pop(0)
                node = self.enemies_right.pop(0)
                self.enemy_fall(node, True)
                self.move_enemies()
                self.add_enemy()
                self.handle_success()


        self.print_enemy_lists()
        return True

    def on_key_release (self, key, modifiers):        
        if self.scene_done:
            return
        #self.scene_done = True
        #self.start_gameplay()

    def on_mouse_press(self, x, y, d, f):
        if self.scene_done:
            return
        self.scene_done = True
        #self.start_gameplay()
    
    def start_gameplay(self):
        scn = gameplay.CreateGameScene()
        cocos.director.director.replace(SlideInRTransition(scn, duration=0.5))
        #cocos.director.director.run(scn)

def CreateScene():
    cr = CreatorReader()
    scene = cr.LoadScene(os.path.join("Scene", "helloworld.ccreator"))
    scene.add(GamePlayScreen(scene))
    return scene